package com.loafle.commons.rpc.registry;

import java.lang.reflect.InvocationTargetException;
import java.lang.reflect.Method;
import java.lang.reflect.Type;
import java.util.HashMap;
import java.util.Map;

import com.loafle.commons.rpc.RPCException;
import com.loafle.commons.rpc.protocol.RPCRegistryCodec;

/**
 * RPCRegistry
 */
public class RPCRegistry implements RPCInvoker {
  private Map<String, RPCService> services;

  public RPCRegistry() {
    this.services = new HashMap<>();
  }

	public RPCService registerService(Object receiver) throws RPCException {
    if (null == receiver) {
      throw new RPCException("receiver is null");
    }
		return this.registerService(receiver, receiver.getClass().getName());
	}

	public RPCService registerService(Object receiver, String name) throws RPCException {
    if (null == receiver) {
      throw new RPCException("receiver is null");
    }

		RPCService rService = new RPCService();
		rService.instance = receiver;
		this.services.put(name, rService);

		return rService;
	}

	public void registerService(Object receiver, Map<String, Method> methods) throws RPCException {
		this.registerService(receiver, receiver.getClass().getName(), methods);
	}

	public void registerService(Object receiver, String name, Map<String, Method> methods) throws RPCException {
		RPCService rService = this.registerService(receiver, name);

		methods.forEach((methodName, method)->{
			rService.registerMethod(method, methodName);
		});
	}

	@Override
	public boolean hasMethod(String method) {
    String[] sm = method.split("\\.");

    if (null == sm || 2 != sm.length) {
      return false;
    }
    RPCService rService = this.services.get(sm[0]);
    if (null == rService) {
      return false;
    }
    RPCServiceMethod rServiceMethod = rService.getServiceMethod(sm[1]);
    if (null == rServiceMethod) {
      return false;
    }
		return true;
	}
	@Override
	public Object invoke(RPCRegistryCodec codec) throws RPCException {
		String method = codec.method();

    String[] sm = method.split("\\.");

    if (null == sm || 2 != sm.length) {
      throw new RPCException(String.format("method[%s] is not valid", method));
    }
    RPCService rService = this.services.get(sm[0]);
    if (null == rService) {
      throw new RPCException(String.format("service[%s] is not exist", sm[0]));
    }
    RPCServiceMethod rServiceMethod = rService.getServiceMethod(sm[1]);
    if (null == rServiceMethod) {
      throw new RPCException(String.format("method[%s] of service[%s] is not exist", sm[1], sm[0]));
    }

		Type[] parameterTypes = rServiceMethod.getParameterTypes();
		Object[] params = codec.params(parameterTypes);

    try {
      return rServiceMethod.method.invoke(rService.instance, params);
    } catch (IllegalAccessException | IllegalArgumentException | InvocationTargetException e) {
      throw new RPCException(String.format("Cannot invoke method[%s] with params[%s]", method, params), e);
    }
	}

  public static class RPCService {
    private Object instance;
    private Map<String, RPCServiceMethod> methods;

    RPCService() {
      this.methods = new HashMap<>();
    }

    public Object getInstance() {
      return this.instance;
    }

		public void registerMethod(Method method) {
			this.registerMethod(method, method.getName());
		}

		public void registerMethod(Method method, String name) {
      RPCServiceMethod rServiceMethod = new RPCServiceMethod();
      rServiceMethod.method = method;
      rServiceMethod.returnType = method.getReturnType();
      rServiceMethod.parameterTypes = method.getGenericParameterTypes();

			this.methods.put(name, rServiceMethod);
		}

    public RPCServiceMethod getServiceMethod(String name) {
      return this.methods.get(name);
    }
  }

  public static class RPCServiceMethod {
    private Method method;
    private Class<?> returnType;
    private Type[] parameterTypes;

    RPCServiceMethod() {
    }

    public Method getMethod() {
      return this.method;
    }

    public Class<?> getReturnType() {
      return this.returnType;
    }

    public Type[] getParameterTypes() {
      return this.parameterTypes;
    }
  }
}
